from dataclasses import dataclass
from typing import Protocol

from pandas import DataFrame


# general protocol that defines that every class needs the get_haplotype_id method
class FileIndices(Protocol):
    def get_haplotype_id(
        self, data: DataFrame, ind_id_indx: int, phase_col_indx: int, col_name: str
    ) -> None:
        ...


@dataclass
class HapIBD(FileIndices):
    id1_indx: int = 0
    hap1_indx: int = 1
    id2_indx: int = 2
    hap2_indx: int = 3
    chr_indx: int = 4
    str_indx: int = 5
    end_indx: int = 6
    cM_indx: int = 7

    def get_haplotype_id(
        self, data: DataFrame, ind_id_indx: int, phase_col_indx: int, col_name: str
    ) -> None:
        data.loc[:, col_name] = (
            data[ind_id_indx] + "." + data[phase_col_indx].astype(str)
        )

    def __str__(self):
        """Custom string message used for debugging"""
        return f"HapIBD: id1_index={self.id1_indx}, id2_index={self.id2_indx}, haplotype_1_index={self.hap1_indx}, haplotype_2_index={self.hap2_indx}, chromosome_index={self.chr_indx}, start_position_index={self.str_indx}, end_position_index={self.end_indx}, centimorgan_index={self.cM_indx}"  # noqa: E501


@dataclass
class Germline(FileIndices):
    id1_indx: int = 0
    hap1_indx: int = 1
    id2_indx: int = 2
    hap2_indx: int = 3
    chr_indx: int = 4
    str_indx: int = 5
    end_indx: int = 6
    cM_indx: int = 10
    unit: int = 11

    def get_haplotype_id(
        self, data: DataFrame, ind_id_indx: int, phase_col_indx: int, col_name: str
    ) -> None:
        data.loc[:, col_name] = data[phase_col_indx]

    def __str__(self):
        """Custom string message used for debugging"""
        return f"Germline: id1_index={self.id1_indx}, id2_index={self.id2_indx}, haplotype_1_index={self.hap1_indx}, haplotype_2_index={self.hap2_indx}, chromosome_index={self.chr_indx}, start_position_index={self.str_indx}, end_position_index={self.end_indx}, centimorgan_index={self.cM_indx}, unit_index={self.unit}"  # noqa: E501


@dataclass
class iLASH(FileIndices):
    id1_indx: int = 0
    hap1_indx: int = 1
    id2_indx: int = 2
    hap2_indx: int = 3
    chr_indx: int = 4
    str_indx: int = 5
    end_indx: int = 6
    cM_indx: int = 9

    def get_haplotype_id(
        self, data: DataFrame, ind_id_indx: int, phase_col_indx: int, col_name: str
    ) -> None:
        data.loc[:, col_name] = data[phase_col_indx]

    def __str__(self):
        """Custom string message used for debugging"""
        return f"iLASH: id1_index={self.id1_indx}, id2_index={self.id2_indx}, haplotype_1_index={self.hap1_indx}, haplotype_2_index={self.hap2_indx}, chromosome_index={self.chr_indx}, start_position_index={self.str_indx}, end_position_index={self.end_indx}, centimorgan_index={self.cM_indx}"  # noqa: E501


@dataclass
class Rapid(FileIndices):
    id1_indx: int = 1
    hap1_indx: int = 3
    id2_indx: int = 2
    hap2_indx: int = 4
    chr_indx: int = 0
    cM_indx: int = 7
    str_indx: int = 5
    end_indx: int = 6

    def get_haplotype_id(
        self, data: DataFrame, ind_id_indx: int, phase_col_indx: int, col_name: str
    ) -> None:
        data.loc[:, col_name] = (
            data[ind_id_indx] + "." + data[phase_col_indx].astype(str)
        )

    def __str__(self):
        """Custom string message used for debugging"""
        return f"Rapid: id1_index={self.id1_indx}, id2_index={self.id2_indx}, haplotype_1_index={self.hap1_indx}, haplotype_2_index={self.hap2_indx}, chromosome_index={self.chr_indx}, start_position_index={self.str_indx}, end_position_index={self.end_indx}, centimorgan_index={self.cM_indx}"  # noqa: E501


def create_indices(ibd_file_format: str) -> FileIndices:
    """Factory method to generate the proper file indice object based on the ibd program

    Parameters
    ----------
    ibd_file_format: str
        string indicating what ibd program was used identify IBD segments. EX: hapibd,
        ilash, rapid, and germline. expects this value to be lower case

    Returns
    -------
    FileIndices
        returns an object that conforms to the FileIndices protocol. It will have the
        method getHAPID. It will also have the correct indices for the ibd program

    Raises
    ------
    ValueError
        Raises a value error if the user passes an ibd_file_format that is not hapibd,
        hap-ibd, germline, ilash, rapid
    """
    format_selector = {
        "germline": Germline(),
        "ilash": iLASH(),
        "hapibd": HapIBD(),
        "rapid": Rapid(),
    }

    return format_selector.get(ibd_file_format)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Protocol, Set, TypeVar, Union

T = TypeVar("T", bound="Network")


class Network_Interface(Protocol):
    clst_id: float  # I don't like this attribute being a float but for now it has to remain this way for backwards compatibility
    true_positive_count: int
    true_positive_percent: float
    false_negative_edges: List[int]
    false_negative_count: int
    members: Set[int]
    haplotypes: List[int]
    min_pvalue_str: str = ""
    pvalues: Dict[str, str] = field(default_factory=dict)

    def print_members_list(self) -> str:
        """Returns a string that has all of the members ids separated by space

        Returns
        -------
        str
            returns a string where the members list attribute
            is formatted as a string for the output file. Individuals strings are joined by comma.
        """
        ...


@dataclass
class Network:
    clst_id: float
    true_positive_count: int
    true_positive_percent: float
    false_negative_edges: List[int]
    false_negative_count: int
    members: Union[Set[int], Set[str]]
    haplotypes: Union[List[int], List[str]]
    min_pvalue_str: str = ""
    pvalues: Dict[str, Dict[str, Any]] = field(default_factory=dict)

    def print_members_list(self) -> str:
        """Returns a string that has all of the members ids separated by space

        Returns
        -------
        str
            returns a string where the members list attribute
            is formatted as a string for the output file. Individuals strings are joined by comma.
        """
        return ", ".join(list(map(str, self.members)))

    def __lt__(self, comp_class: T) -> bool:
        """Override the less than method so that objects can be sorted in
        ascending numeric order based on cluster id.

        Parameters
        ----------
        comp_class : Network
            Network object to compare.

        Returns
        -------
        bool
            returns true if the self cluster id is less than the
            comp_class cluster id.
        """

        return self.clst_id < comp_class.clst_id



from dataclasses import dataclass
from typing import List

from drive.factory import factory_register
from drive.log import CustomLogger
from drive.models import Data_Interface, Network_Interface

logger = CustomLogger.get_logger(__name__)


@dataclass
class NetworkWriter:
    """Class that is responsible for creating the *_networks.
    txt file from the information provided"""

    name: str = "NetworkWriter plugin"

    @staticmethod
    def _form_header(phenotypes: List[str]) -> str:
        """Method that will form the header line for the output file

        Parameters
        ----------
        phenotypes : List[str]
            list of all the phenotypes from the Data.carriers attribute

        Returns
        -------
        str
            returns the header string for the file
        """

        # making a string for the initial first few columns
        header_str = "clstID\tn.total\tn.haplotype\ttrue.positive.n\ttrue.positive\tfalst.postive\tIDs\tID.haplotype"  # noqa: E501

        if not phenotypes:
            return header_str + "\n"
        else:
            # We need to add columns for the min pvalue descriptions
            header_str += "\tmin_pvalue\tmin_phenotype\tmin_phenotype_description"
            # for each phenotype we are going to create 4 columns for the number
            # of cases in the network, The case ids in the network the number of
            # excluded individuals in the network, and the pvalue for the phenotype
            for column in phenotypes:
                header_str += f"\t{column + '_case_count_in_network'}\t{column + 'cases_in_network'}\t{column + '_excluded_count_in_network'}\t{column + 'excluded_in_network'}\t{column + '_pvalue'}"  # noqa: E501

            return header_str + "\n"

    @staticmethod
    def _create_network_info_str(
        network: Network_Interface, phenotypes: List[str]
    ) -> str:
        """create a string that has all the information from the network
        such as cluster id, member count, pvalues, etc...

        Parameters
        ----------
        network : Network_Interface
            network object that hsas all the per network
            information from the clusters such as cluster
            ids, networks, pvalues, etc...

        phenotypes : List[str]
            list of the phenotypes provided the program.

        Returns
        -------
        str
            returns a string formatted for the output file
        """
        # fill in the initial few columns of the output string
        output_str = f"{network.clst_id}\t{len(network.members)}\t{len(network.haplotypes)}\t{network.true_positive_count}\t{network.true_positive_percent:.4f}\t{network.false_negative_count}\t{','.join(network.members)}\t{','.join(network.haplotypes)}"  # noqa: E501

        if not phenotypes:
            return output_str + "\n"
        else:
            output_str += f"\t{network.min_pvalue_str}"

            for phenotype in phenotypes:
                output_str += f"\t{network.pvalues[phenotype]}"

            return output_str + "\n"

    def analyze(self, **kwargs) -> None:
        """main function of the plugin that will create the
        output path and then use helper functions to write
        information to a file"""

        data: Data_Interface = kwargs["data"]

        # creating the full output path for the output file
        network_file_output = data.output_path.parent / (
            data.output_path.name + ".drive_networks.txt"
        )  # noqa: E501

        logger.debug(
            f"The output in the network_writer plugin is being written to: {network_file_output}"  # noqa: E501
        )

        # we are going to pull out the phenotypes into a list so that we
        # are guarenteed to maintain order as we are creating the rows
        phenotypes = list(data.carriers.keys())

        with open(network_file_output, "w", encoding="utf-8") as networks_output:
            header_str = NetworkWriter._form_header(phenotypes)
            # iterate over each network and pull out the appropriate
            # information into strings
            _ = networks_output.write(header_str)

            for network in data.networks:
                network_info_str = NetworkWriter._create_network_info_str(
                    network, phenotypes
                )

                networks_output.write(network_info_str)


def initialize() -> None:
    factory_register("network_writer", NetworkWriter)



# File will have different Interfaces used in the program as well as different class for types. This will simplify the imports and allow me to reduce coupling between modules

from collections import namedtuple
from dataclasses import dataclass, field
from typing import Protocol

from pandas import DataFrame

# namedtuple that will contain information about the gene being run
Genes = namedtuple("Genes", ["chr", "start", "end"])


# interface for the filter object
@dataclass
class Filter(Protocol):
    ibd_vs: DataFrame = field(default_factory=DataFrame)
    ibd_pd: DataFrame = field(default_factory=DataFrame)



from pathlib import Path
from typing import Dict, Union

from pandas import read_csv


def load_phenotype_descriptions(
    phecode_desc_file: Union[Path, str]
) -> Dict[str, Dict[str, str]]:
    """Function that will load the phecode_description file
    and then turn that into a dictionary

    Parameters
    ----------
    phecode_desc_file : str
        descriptions of each phecode

    Returns
    -------
    Dict[str, Dict[str, str]]
        returns a dictionary where the first key is the
        phecode and value is a dictionary where the inner key
        is 'phenotype' and the value is the descriptions
    """

    desc_df = read_csv(phecode_desc_file, sep="\t", usecols=["phecode", "phenotype"])

    # making sure that the phecode keys are a string
    desc_df.phecode = desc_df.phecode.astype(str)

    # converting the dataframe into a dictionar
    return desc_df.set_index("phecode").T.to_dict()



"""Module to faciliate the user in parsing the phenotype file by incorporating multiple 
ecodings, separators, and by handling multiple errors."""

from logging import Logger
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, TypeVar, Union

import pandas as pd

from drive.log import CustomLogger

logger: Logger = CustomLogger.get_logger(__name__)

# creating a type annotation for the PhenotypeFileParser class
T = TypeVar("T", bound="PhenotypeFileParser")

PhenotypeInfo = TypeVar("PhenotypeInfo", bound=dict)


class PhenotypeFileParser:
    """Parser used to read in the phenotype file. This will allow use to account for
    different delimiters in files as well as catch errors."""

    def __init__(
        self, filepath: Union[Path, str], phenotype_name: Optional[str] = None
    ) -> None:
        """Initialize the PhenotypeFileParser class.

        Parameters
        ----------
        filepath : Path | str
            filepath to the phenotype file that has case control status for individuals

        phenotype_name : str
            Phenotype name that can be used specify a specific column in a
            phenotype matrix if the user only wants ot focus on 1 phenotype.

        Raises
        ------
        FileNotFoundError
        """
        self.specific_phenotype: str = phenotype_name
        # we are going to make sure the filepath variable is a
        # PosixPath
        filepath = Path(filepath)

        # now we are going to try to create an attribute for
        # the input file
        if not filepath.exists():
            raise FileNotFoundError(f"The file {filepath} was not found")
        else:
            self.file: Path = filepath

    def __enter__(self) -> T:
        """Open the input file. Method determines the appropriate file type and open
        the file. Method is called automatically by the context manager.

        Raises
        ------
        pd.errors.ParserError
            Raised if the program encounters any error while trying to read in the
            phenotype matrix using pd.read_csv
        """
        try:
            pheno_df = pd.read_csv(
                self.file,
                sep="\t",
                na_values=["na", "n/a", "-1", "-1.0", " ", "", "NA", "N/A"],
            ).fillna(-1)

        except pd.errors.ParserError as e:
            logger.critical(e)
            logger.critical(
                f"Encountered the following error while trying to read in the phenotype matrix: {self.file}"  # noqa: E501
            )
            raise pd.errors.ParserError(
                f"Encountered the following error while trying to read in the phenotype matrix: {self.file}"  # noqa: E501
            )

        self.phenotype_df = pheno_df

        logger.verbose(
            f"Reading in {self.phenotype_df.shape[1] - 1} phecodes for {self.phenotype_df.shape[0]} individuals"  # noqa: E501
        )  # noqa: E501

        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Close the resource when it is not longer being used. Used by the context
        manager."""
        logger.verbose("Finished reading in individuals from the phenotype file")

    def _generate_columns_to_keep(self) -> List[str]:
        """determines which columns in the phecode file to use

        Returns
        -------
        List[str]
        returns a list of strings. If the user specified a specific phecode then
        it returns every phecode otherwise it returns a single phecode in a list

        Raises
        ------
        ValueError
            if the specified PheCode is not found in the phenotype file then an error is
            logged and raised
        """
        if (
            self.specific_phenotype is not None
            and self.specific_phenotype not in self.phenotype_df.columns
        ):  # noqa: E501
            logger.critical(
                f"The phenotype {self.specific_phenotype} was not found in the phenotype file"  # noqa: E501
            )  # noqa: E501
            raise ValueError(
                f"The phenotype {self.specific_phenotype} was not found in the phenotype file"  # noqa: E501
            )  # noqa: E501
        elif self.specific_phenotype is not None:
            return [self.specific_phenotype]
        else:
            return list(self.phenotype_df.columns[1:])

    def _process_matrix(
        self, columns: List[str]
    ) -> Tuple[Dict[str, Dict[str, Set[str]]], List[str]]:
        """Function that will generate a dictionary where the keys are
        phenotypes and the values are lists of the cases/exclusions/controls

        Parameters
        ----------
        columns : List[str]
            list of columns from the phenotype matrix. If the user provides
            a specific phenotype then this list will be of size one, otherwise it will
            be the size of the number of phecodes in the file.

        Returns
        -------
        Tuple[Dict[str, Dict[str, Set[str]]], List[str]]
            returns a tuple with three elements. The first element is a
            dictionary where the keys are phenotypes. Values are
            dictionaries where the keys are 'cases' or 'controls' or
            'excluded' and values are list of ids. The second element is
            a dictionary that maps the index of the phenotype in the
            header line to the phenotype name. The third element is the
            separator string
        """

        # we need to create the dictionary that will have the case/control/exclusion
        # ids for each phecode
        phenotype_dict = {phecode: None for phecode in columns}

        # we need to create a dictionary that will map the index of the phenotype

        # we ultimately are going to just want to filter this grids series
        grids = self.phenotype_df.iloc[:, 0].astype(str)

        # We will pull out values for each phenotype to determine the cases/controls/
        # exclusions
        for phecode_name, phenotyping_status in self.phenotype_df.loc[
            :, columns
        ].items():
            cases = set(
                grids[phenotyping_status[phenotyping_status == 1].index].unique()
            )

            controls = set(
                grids[phenotyping_status[phenotyping_status == 0].index].unique()
            )

            exclusions = set(
                grids[phenotyping_status[phenotyping_status == -1].index].unique()
            )

            phenotype_dict[phecode_name] = {
                "cases": cases,
                "controls": controls,
                "excluded": exclusions,
            }

        logger.info(
            f"For PheCode, {phecode_name}, identified {len(phenotype_dict[phecode_name]['cases'])} cases, {len(phenotype_dict[phecode_name]['controls'])} controls, and {len(phenotype_dict[phecode_name]['excluded'])} exclusions"
        )  # noqa: E501
        return phenotype_dict, grids.values.tolist()

    def parse_cases_and_controls(
        self,
    ) -> Tuple[Dict[str, Dict[str, Set[str]]], Dict[int, str]]:
        """Generate a list for cases, controls, and excluded individuals.

        Returns
        -------
        Tuple[Dict[str, Dict[str, Set[str]]], List[str]]
            returns a tuple where the first element is a dictionary where
            the keys are the phenotypes and the values are dictionary of
            the case/controls/excluded individuals lists. The second
            element is a list of all grids from the file to be used as a
            cohort
        """

        cols_to_keep = self._generate_columns_to_keep()

        phenotyping_dictionary, cohort_ids = self._process_matrix(cols_to_keep)

        return phenotyping_dictionary, cohort_ids



from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple

from numpy import float64
from scipy.stats import binomtest

from drive.factory import factory_register
from drive.log import CustomLogger
from drive.models import Data_Interface, Network_Interface

logger = CustomLogger.get_logger(__name__)


@dataclass
class Pvalues:
    """Class that is responsible for determining the pvalues for each network"""

    name: str = "Pvalue plugin"

    @staticmethod
    def _determine_pvalue(
        phenotype: str,
        phenotype_percent: float,
        carriers_count: int,
        network_size: int,
    ) -> float64:
        """Function that will determine the pvalue for each network

        Parameters
        ----------
        phenotype : str
            id for the phenotype

        phenotype_percent : float
            float that tells what the prevalence of the phenotype in the
            cohort population is

        carriers_count : int
            number of carriers in the network

        network_size : int
            size of the network

        Returns
        -------
        float
            Returns the calculated pvalue
        """
        # the probability is 1 if the carrier count is zero because it is chances of
        # finding 0 or higher which is everyone
        if carriers_count == 0:
            logger.debug(f"carrier count = 0 therefore pvalue for {phenotype} = 1")
            return 1

        result = binomtest(carriers_count - 1, network_size, phenotype_percent)

        pvalue = result.pvalue

        logger.debug(f"pvalue for {phenotype} = {pvalue}")

        return pvalue

    @staticmethod
    def _determine_phenotype_frequency(
        phecode: str, phenotype_counts: Dict[str, List[str]]
    ) -> float:
        """calculate the phenotype frequency in the cohort

        Parameters
        ----------
        phecode : str
            phecode id. This value will only be used for a logging statement

        phenotype_counts : Dict[str, List[str]]
            Dictionary that has list for individuals who are
            cases, controls, or exclusions.

        Returns
        -------
        float
            returns the phenotype frequency as a float
        """

        # We need to first check if there are even controls in the file
        phenotype_frequency = len(phenotype_counts.get("cases")) / (
            len(phenotype_counts.get("controls")) + len(phenotype_counts.get("cases"))
        )

        logger.verbose(
            f"For PheCode, {phecode}, {len(phenotype_counts.get('cases'))} cases and {len(phenotype_counts.get('controls'))} were identified giving a phenotype frequency of {phenotype_frequency}"  # noqa: E501
        )

        return phenotype_frequency

    @staticmethod
    def _get_carrier_count_in_network(
        phenotype_counts: Dict[str, List[str]], network: Network_Interface
    ) -> int:
        """Determine the number of individual cases that are
        also in the network

        Parameters
        ----------
        phenotype_counts : Dict[str, List[str]]
            Dictionary that has list for individuals who are
            cases, controls, or exclusions.

        network : Network_Interface
            Network object with information about members of
            the networks and what haplotypes are in the network

        Returns
        -------
        int
            returns the number of carriers in the network
        """
        # determine the number of carriers in the network

        return len(network.members.intersection(phenotype_counts.get("cases")))

    @staticmethod
    def _get_carriers_in_network(
        phenotype_counts: Dict[str, List[str]], network: Network_Interface
    ) -> str:
        """Generate a set of cases that are in the network

        Parameters
        ----------
        phenotype_counts : Dict[str, List[str]]
            Dictionary that has list for individuals who are
            cases, controls, or exclusions.

        network : Network_Interface
            Network object with information about members of
            the networks and what haplotypes are in the network

        Returns
        -------
        str
            returns a string of individuals that are in both the phenotype_counts
            case set and the network members set. If this interection is empty
            then it returns a str "N/A"
        """
        # determine the number of carriers in the network
        case_individuals = network.members.intersection(phenotype_counts.get("cases"))

        if case_individuals:
            return ", ".join(
                network.members.intersection(phenotype_counts.get("cases"))
            )  # noqa: E501
        else:
            return "N/A"

    @staticmethod
    def _get_exclusions_in_network(
        phenotype_counts: Dict[str, List[str]], network: Network_Interface
    ) -> str:
        """Generate a set of cases that are in the network

        Parameters
        ----------
        phenotype_counts : Dict[str, List[str]]
            Dictionary that has list for individuals who are
            cases, controls, or exclusions.

        network : Network_Interface
            Network object with information about members of
            the networks and what haplotypes are in the network

        Returns
        -------
        str
            returns a string of individuals that are in both the phenotype_counts
            exclusion set and the network members set. If this interection is
            empty then it returns a str "N/A"
        """
        # determine the number of carriers in the network
        excluded_individuals = network.members.intersection(
            phenotype_counts.get("excluded")
        )  # noqa: E501

        if excluded_individuals:
            return ", ".join(
                network.members.intersection(phenotype_counts.get("excluded"))
            )  # noqa: E501
        else:
            "N/A"

    def _remove_exclusions(
        phenotype_counts: Dict[str, List[str]], network: Network_Interface
    ) -> Tuple[int, int]:
        """determine size of network after removing excluded
        individuals

        Parameters
        ----------
        phenotype_counts : Dict[str, List[str]]
            Dictionary that has list for individuals who are
            cases, controls, or exclusions.

        network : Network_Interface
            Network object with information about members of
            the networks and what haplotypes are in the network

        Returns
        -------
        Tuple[int, int]
            returns the number of individuals in the network,
            not counting those individuals classified as
            excluded in the phenotype_counts dictionary. Also
            returns the number of individuals excluded. Also returns the
        """

        return (
            len(network.members.difference(phenotype_counts.get("excluded"))),
            len(network.members.intersection(phenotype_counts.get("excluded"))),
        )

    def _gather_network_information(
        self,
        network: Network_Interface,
        cohort_carriers: Dict[str, Dict[str, List[str]]],
    ) -> tuple[str, str, Dict[str, str]]:
        """Determine information about how many
        carriers are in each network, the percentage, the IIDs of
        the carriers in the network, and use this to calculate the
        pvalue for the network. The function keeps track of the
        smallest non-zero pvalue and returns it or NA

        Parameters
        ----------

        network : Network
            Network object attributes for iids, pairs, and haplotypes

        cohort_carriers : Dict[str, Dict[str, List[str]]]
            dictionary where the keys are phenotype ids and the values are a dictionary
            with the list of cases/controls/exclusions

        Returns
        -------
        Tuple[str, str, Dict[str, str]]
            returns a tuple where the first element is the string of the
            minimum phenotype code. The second value is the description
            of the minimum phenotype. The third value is a dictionary
            that for each phenotype has a string with the number of
            cases, number of excluded individuals, and the p-values.
        """
        # setting null values for the output string and the cur_min_phecode string.
        # Also setting a value of 1 for the cur_min_pvalue because this will be the
        # largest pvalue possible

        cur_min_pvalue = 1

        cur_min_phenotype = ""

        phenotype_pvalues = {}

        # iterating over each phenotype
        for phenotype, phenotype_counts in cohort_carriers.items():
            # if there are no controls then we can't do the
            # stats and should just return a string of N/A for all values
            if len(phenotype_counts.get("controls")) == 0:
                phenotype_pvalues[phenotype] = "N/A\tN/A\tN/A\tN/A"
            else:
                phenotype_freq = Pvalues._determine_phenotype_frequency(
                    phenotype, phenotype_counts
                )

                num_carriers_in_network = Pvalues._get_carrier_count_in_network(
                    phenotype_counts, network
                )

                carrier_str = Pvalues._get_carriers_in_network(
                    phenotype_counts, network
                )

                exclusion_str = Pvalues._get_exclusions_in_network(
                    phenotype_counts, network
                )

                (
                    network_size_after_exclusions,
                    excluded_count,
                ) = Pvalues._remove_exclusions(phenotype_counts, network)

                # calling the sub function that determines the pvalue
                pvalue: float = Pvalues._determine_pvalue(
                    phenotype,
                    phenotype_freq,
                    num_carriers_in_network,
                    network_size_after_exclusions,
                )

                # Next two lines create the string and then concats it to the output_str

                phenotype_str = f"{num_carriers_in_network}\t{carrier_str}\t{excluded_count}\t{exclusion_str}\t{pvalue}"  # noqa: E501

                phenotype_pvalues[phenotype] = phenotype_str

                # Now we will see if the phecode is lower then the cur_min_pvalue.
                # If it is then we will change the cur_min_pvalue and we will
                # update the cur_min_phecode
                if pvalue < cur_min_pvalue and pvalue != 0:
                    cur_min_pvalue = pvalue

                    cur_min_phenotype = phenotype

        # if a minimum phecode is identified then we need to create a string, otherwise
        # we use N/A's
        if not cur_min_phenotype:
            cur_min_phenotype = "N/A"
            cur_min_pvalue = "N/A"

        # return the pvalue_output string first and either a tuple of N/As or the min
        # pvalue/min_phecode
        return cur_min_pvalue, cur_min_phenotype, phenotype_pvalues

    @staticmethod
    def _get_descriptions(
        phecode_description: dict[str, dict[str, str]], min_phecode: str
    ) -> str:
        """Method to get the description for the minimum phecode
        Parameters
        ----------
        phecode_descriptions : dict[str, dict[str, str]]
            dictionary with descriptions of each phecode. The outer key is the
            phenotype id. The inner key is the string phenotype and the value is a
            string that describes the phenotype

        min_phecode : str
            minimum phecode string

        Returns
        -------
        str
            returns a string that has the phecode description
        """
        # getting the description
        desc_dict = phecode_description.get(min_phecode, {})

        # getting the phenotype string if key exists,
        # otherwise returns an empty string
        logger.debug(f"description_str = {desc_dict.get('phenotype', 'N/A')}")

        return desc_dict.get("phenotype", "N/A")

    def analyze(self, **kwargs) -> None:
        # this is the DataHolder model. We will use the networks, the
        # affected_inds, and the phenotype_prevalances attribute
        data: Data_Interface = kwargs["data"]

        if data.carriers:
            for network in data.networks:
                # Determining the pvalues for the network
                (
                    min_pvalue_str,
                    min_phenotype_code,
                    phenotype_pvalues,
                ) = self._gather_network_information(network, data.carriers)

                min_phecode_description = self._get_descriptions(
                    data.phenotype_descriptions, min_phenotype_code
                )

                # create an attribute that has all the strings for what is the
                # min_pvalue, the min_phenotype_code, and the description of the
                # phenotype
                network.min_pvalue_str = (
                    f"{min_pvalue_str}\t{min_phenotype_code}\t{min_phecode_description}"
                )

                # adding the dictionary of phenotype values to the attribute
                # pvalues
                network.pvalues = phenotype_pvalues


def initialize() -> None:
    factory_register("pvalues", Pvalues)



from typing import Any, Callable, Protocol


class PluginNotFound(Exception):
    """
    Error that is raised if the user tries to load a plugin
    that is not there
    """

    def __init__(self, plugin_type: str) -> None:
        super().__init__(
            f"Plugin, {plugin_type} not found in the plugins folder. Make sure you loaded the plugin into the json file."  # noqa: E501
        )


class AnalysisObj(Protocol):
    """Interface defining of an analysis object"""

    def analyze(self, **kwargs) -> Any:
        """
        Method that will analyze the inputs according to the purpose
        of the plugin
        """


analyze_obj_creation_funcs: dict[str, Callable[..., AnalysisObj]] = {}


def register(plugin_name: str, creation_func: Callable[..., AnalysisObj]) -> None:
    """registers the AnalysisObj plugin"""
    analyze_obj_creation_funcs[plugin_name] = creation_func


def unregister(plugin_name: str) -> None:
    """function that will unregister the plugin

    Parameters
    ----------
    plugin_name : str
        name of the plugin to be loaded. This name will be in json file
    """
    analyze_obj_creation_funcs.pop(plugin_name, None)


def create(arguments: dict[str, Any]) -> AnalysisObj:
    args_copy = arguments.copy()

    plugin_type: str = args_copy.pop("name")

    try:
        creation_func = analyze_obj_creation_funcs[plugin_type]
        return creation_func(**args_copy)
    except KeyError:
        raise PluginNotFound(plugin_type)  # pylint: disable=raise-missing-from



"""File used to load in the different plugins"""

import importlib
from typing import Protocol


class PluginInterface(Protocol):
    """Interface that will define how a plugin looks like"""

    @staticmethod
    def initialize() -> None:
        """Method that will initialize the plugin"""


def import_module(name: str) -> PluginInterface:
    """Import the plugin so it can be initialized

    Parameters
    ----------
    name : str
        name of the plugin

    Returns
    -------
    PluginInterface
    """
    return importlib.import_module(name)  # type: ignore


def load_plugins(plugins: list[str]) -> None:
    """Calls the initialize method for each plugin"""
    for plugin_name in plugins:
        plugin = import_module(plugin_name)

        plugin.initialize()



import argparse
import json
import re
from datetime import datetime
from importlib.metadata import version
from pathlib import Path

from rich_argparse import RichHelpFormatter

import drive.factory as factory
from drive.cluster import ClusterHandler, cluster
from drive.filters import IbdFilter
from drive.log import CustomLogger
from drive.models import Data, Genes, create_indices
from drive.utilities.callbacks import CheckInputExist
from drive.utilities.parser import PhenotypeFileParser, load_phenotype_descriptions


def find_json_file() -> Path:
    """Method to find the default config file if the user does not provide one

    Returns
    -------
    Path
        returns the path to the json file

    Raises
    ------
    FileNotFoundError
        Raises a FileNotFoundError if the program can not locate a json file and the
        user does not provide the path to a file
    """

    src_dir = Path(__file__).parent

    config_path = src_dir / "config.json"

    if not config_path.exists():
        raise FileNotFoundError(
            f"Expected the user to either pass a configuration file path or for the config.json file to be present in the program root directory at {config_path}."  # noqa: E501
        )

    return config_path


def split_target_string(chromo_pos_str: str) -> Genes:
    """Function that will split the target string provided by the user.

    Parameters
    ----------
    chromo_pos_str : str
        String that has the region of interest in base pairs.
        This string will look like 10:1234-1234 where the
        first number is the chromosome number, then the start
        position, and then the end position of the region of
        interest.

    Returns
    -------
    Genes
        returns a namedtuple that has the chromosome number,
        the start position, and the end position

    Raises
    ------
    ValueError
        raises a value error if the string was formatted any
        other way than chromosome:start_position-end_position.
        Also raises a value error if the start position is
        larger than the end position
    """
    split_str = re.split(":|-", chromo_pos_str)

    if len(split_str) != 3:
        error_msg = f"Expected the gene position string to be formatted like chromosome:start_position-end_position. Instead it was formatted as {chromo_pos_str}"  # noqa: E501

        raise ValueError(error_msg)

    integer_split_str = [int(value) for value in split_str]

    if integer_split_str[1] > integer_split_str[2]:
        raise ValueError(
            f"expected the start position of the target string to be <= the end position. Instead the start position was {integer_split_str[1]} and the end position was {integer_split_str[2]}"  # noqa: E501
        )

    return Genes(*integer_split_str)


def main() -> None:
    parser = argparse.ArgumentParser(
        description=" Distant Relatedness for Identification and Variant Evaluation (DRIVE) is a novel approach to IBD-based genotype inference used to identify shared chromosomal segments in dense genetic arrays",
        formatter_class=RichHelpFormatter,
    )

    parser.add_argument(
        "--input",
        "-i",
        type=Path,
        help="IBD input file from ibd detection software",
        required=True,
        action=CheckInputExist,
    )

    parser.add_argument(
        "--format",
        "-f",
        default="hapibd",
        type=str,
        help="IBD program used to detect segments. Allowed values are hapibd, ilash, germline, rapid. Program expects for value to be lowercase. (default: %(default)s)",
        choices=["hapibd", "ilash", "germline", "rapid"],
    )

    parser.add_argument(
        "--target",
        "-t",
        type=str,
        help="Target region or position, chr:start-end or chr:pos",
        required=True,
    )

    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        help="output file prefix. The program will append .drive_networks.txt to the filename provided",
        required=True,
    )

    parser.add_argument(
        "--min-cm",
        "-m",
        default=3,
        type=int,
        help="minimum centimorgan threshold. The program expects this to be an integer value. (default: %(default)s)",
    )

    parser.add_argument(
        "--step",
        "-k",
        default=3,
        type=int,
        help="Minimum required number of steps for the community walktrap algorithm.(default: %(default)s)",
    )

    parser.add_argument(
        "--max-recheck",
        default=5,
        type=int,
        help="Maximum number of times to re-perform the clustering. This value will not be used if the flag --no-recluster is used.(default: %(default)s)",  # noqa: E501
    )

    parser.add_argument(
        "--cases",
        "-c",
        type=Path,
        help="A file containing individuals who are cases. This file expects for there to be two columns. The first column will have individual ids and the second has status where cases are indicated by a 1, control are indicated by a 0, and exclusions are indicated by NA.",  # noqa: E501
        action=CheckInputExist,
    )

    parser.add_argument(
        "--segment-overlap",
        default="contains",
        choices=["contains", "overlaps"],
        type=str,
        help="Indicates if the user wants the gene to contain the whole target region or if it just needs to overlap the segment. (default: %(default)s)",  # noqa: E501
    )

    parser.add_argument(
        "--descriptions",
        "-d",
        type=Path,
        help="tab delimited text file that has descriptions for each phecode. this file should have two columns called phecode and phenotype",  # noqa: E501
    )

    parser.add_argument(
        "--max-network-size",
        default=30,
        type=int,
        help="maximum network size allowed if the user has allowed the recluster option. (default: %(default)s)",
    )

    parser.add_argument(
        "--min-connected-threshold",
        default=0.5,
        type=float,
        help="minimum connectedness ratio required for the network. (default: %(default)s)",
    )

    parser.add_argument(
        "--min-network-size",
        default=2,
        type=int,
        help="This argument sets the minimun network size that we allow. All networks smaller than this size will be filtered out. If the user wishes to keep all networks they can set this to 0. (default: %(default)s)",  # noqa: E501
    )

    parser.add_argument(
        "--segment-distribution-threshold",
        default=0.2,
        type=float,
        help="Threshold to filter the network length to remove hub individuals. (default: %(default)s)",
    )

    parser.add_argument(
        "--phenotype-name",
        default=None,
        type=str,
        help="If the user wishes to only run 1 phenotype from a file with multiple phenotypes they can prioritize a column that has the phenotype name. The name must match with the column.",
    )

    parser.add_argument(
        "--hub-threshold",
        default=0.01,
        type=float,
        help="Threshold to determine what percentage of hubs to keep. (default: %(default)s)",
    )

    parser.add_argument(
        "--json-config",
        "-j",
        default=None,
        type=Path,
        help="path to the json config file",
    )

    parser.add_argument(
        "--recluster",
        type=bool,
        default=True,
        action=argparse.BooleanOptionalAction,
        help="whether or not the user wishes the program to automically recluster based on things like hub threshold, max network size and how connected the graph is. ",  # noqa: E501
    )

    parser.add_argument(
        "--verbose",
        "-v",
        default=0,
        help="verbose flag indicating if the user wants more information",
        action="count",
    )

    parser.add_argument(
        "--log-to-console",
        default=False,
        help="Optional flag to log to only the console or also a file",
        action="store_true",
    )

    parser.add_argument(
        "--log-filename",
        default="drive.log",
        type=str,
        help="Name for the log output file. (default: %(default)s)",
    )

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s: {version('drive-ibd')}"
    )

    args = parser.parse_args()

    # getting the programs start time
    start_time = datetime.now()

    # We need to make sure that there is a configuration file
    json_config = args.json_config if args.json_config else find_json_file()

    # creating and configuring the logger and then recording user inputs
    logger = CustomLogger.create_logger()

    logger.configure(
        args.output.parent, args.log_filename, args.verbose, args.log_to_console
    )

    # record the input parameters using a method from the logger object that
    # takes the parser as an argument
    logger.record_namespace(args)

    logger.debug(f"Parent directory for log files and output: {args.output.parent}")

    logger.info(f"Analysis start time: {start_time}")
    # we need to load in the phenotype descriptions file to get
    # descriptions of each phenotype
    if args.descriptions:
        logger.verbose(f"Using the phenotype descriptions file at: {args.descriptions}")
        desc_dict = load_phenotype_descriptions(args.descriptions)
    else:
        logger.verbose("No phenotype descriptions provided")
        desc_dict = {}

    # if the user has provided a phenotype file then we will determine case/control/
    # exclusion counts. Otherwise we return an empty dictionary
    if args.cases:
        with PhenotypeFileParser(args.cases, args.phenotype_name) as phenotype_file:
            phenotype_counts, cohort_ids = phenotype_file.parse_cases_and_controls()

            logger.info(
                f"identified {len(phenotype_counts.keys())} phenotypes within the file {args.cases}"  # noqa: E501
            )
    else:
        logger.info(
            "No phenotype information provided. Only the clustering step of the analysis will be performed"  # noqa: E501
        )

        phenotype_counts = {}
        cohort_ids = {}

    indices = create_indices(args.format.lower())

    logger.debug(f"created indices object: {indices}")

    ##target gene region or variant position
    target_gene = split_target_string(args.target)

    logger.debug(f"Identified a target region: {target_gene}")

    filter_obj: IbdFilter = IbdFilter.load_file(args.input, indices, target_gene)

    # choosing the proper way to filter the ibd files
    filter_obj.set_filter(args.segment_overlap)

    filter_obj.preprocess(args.min_cm, cohort_ids)

    # We need to invert the hapid_map dictionary so that the
    # integer mappings are keys and the values are the
    # haplotype string
    hapid_inverted = {value: key for key, value in filter_obj.hapid_map.items()}

    # creating the object that will handle clustering within the networks
    cluster_handler = ClusterHandler(
        args.min_connected_threshold,
        args.max_network_size,
        args.max_recheck,
        args.step,
        args.min_network_size,
        args.segment_distribution_threshold,
        args.hub_threshold,
        hapid_inverted,
        args.recluster,
    )

    networks = cluster(filter_obj, cluster_handler, indices.cM_indx)

    # creating the data container that all the plugins can interact with
    plugin_api = Data(networks, args.output, phenotype_counts, desc_dict)

    logger.debug(f"Data container: {plugin_api}")

    # making sure that the output directory is created
    # This section will load in the analysis plugins and run each plugin
    with open(json_config, encoding="utf-8") as json_config:
        config = json.load(json_config)

        factory.load_plugins(config["plugins"])

        analysis_plugins = [factory.factory_create(item) for item in config["modules"]]

        logger.debug(
            f"Using plugins: {', '.join([obj.name for obj in analysis_plugins])}"
        )

        # iterating over every plugin and then running the analyze and write method
        for analysis_obj in analysis_plugins:
            analysis_obj.analyze(data=plugin_api)

    end_time = datetime.now()

    logger.info(
        f"Analysis finished at {end_time}. Total runtime: {end_time - start_time}"
    )


if __name__ == "__main__":
    main()



import logging
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict

# We are going to configure a specific logging level for
# VERBOSE versus INFO
logging.VERBOSE = logging.INFO - 5


class CustomLogger(logging.getLoggerClass()):
    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        """Initialize the CustomLogger class

        Parameters
        ----------
        name : str
            name of the logger. could be __main__

        level : int
            log level represented as an integer
        """
        super().__init__(name, level)
        # we want a default verbosity level of 15

        logging.addLevelName(logging.VERBOSE, "VERBOSE")

    def verbose(self, msg, *args, **kwargs) -> None:
        """log messages at a verbose level

        Parameters
        ----------
        msg : str
            message provided to the logger
        """
        if self.isEnabledFor(logging.VERBOSE):
            self._log(logging.VERBOSE, msg, args, **kwargs)

    def record_namespace(self, inputs: Namespace) -> None:
        """log the user arguments from the argparse Namespace

        Parameters
        ----------
        inputs : Namespace
            Namespace object that has all of the arguments passed to Argparse parser
        """
        argument_dictionary = vars(inputs)

        current_level = self.getEffectiveLevel()
        # changing the loglevel so that it records the info messages here
        self.setLevel(logging.INFO)

        # going over each value in the kwargs to write to a file
        for parameter, value in argument_dictionary.items():
            self.info(f"{parameter}: {value}")

        # getting the correct log level to reset the logger
        self.setLevel(current_level)

    def record_inputs(self, **kwargs: Dict[str, Any]) -> None:
        """function to record the user arguments that were passed to the
        program. Takes a logger and then a dictionary of the user
        arguments

        Parameters
        ----------
        logger : logging.Logger"""

        current_level = self.getEffectiveLevel()
        # changing the loglevel so that it records the info messages here
        self.setLevel(logging.INFO)

        # going over each value in the kwargs to write to a file
        for parameter, value in kwargs.items():
            self.info(f"{parameter}: {value}")

        # getting the correct log level to reset the logger
        self.setLevel(current_level)

    @staticmethod
    def get_loglevel(loglevel: int) -> int:
        """Function that will return a log level based on the input

        Parameters
        ----------
        loglevel : int
            integer that represents what loglevel the program
            will use. This number will be zero if the user didn't
            pass the verbose flag and will be 1 if the user
            passed the verbose flag.

        Returns
        -------
        int
            returns an integer representing the loglevel in the level_dict
        """
        if loglevel == 0:
            return logging.INFO
        elif loglevel == 1:
            return logging.VERBOSE
        else:
            return logging.DEBUG

    def configure(
        self,
        output: Path,
        filename: str = "test.log",
        verbosity: int = 0,
        to_console: bool = False,
        format_str: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    ) -> None:
        """Function that will configure the logger

        Parameters
        ----------
        logger"""

        filename = output / filename

        self.setLevel(CustomLogger.get_loglevel(verbosity))

        file_formatter = logging.Formatter(format_str)

        # program defaults to log to a file called IBDCluster.log in the
        # output directory
        fh = logging.FileHandler(filename, mode="w")
        fh.setFormatter(file_formatter)
        self.addHandler(fh)

        # If the user selects to also log to console then the program will
        # log information to the stderr
        if to_console:
            stream_formatter = logging.Formatter("%(message)s")

            sh = logging.StreamHandler()
            sh.setFormatter(stream_formatter)
            self.addHandler(sh)

    @staticmethod
    def get_logger(module_name: str, main_name: str = "__main__") -> logging.Logger:
        """Function that will be responsible for getting the logger for modules"""
        return logging.getLogger(main_name).getChild(module_name)

    @staticmethod
    def create_logger(
        logger_name: str = "__main__",
    ) -> logging.Logger:
        """Function that will get the correct logger for the program

        Parameters
        ----------
        loglevel : str
            logging level that the user wants to use. The default level is INFO

        Returns
        -------
        logging.Logger
        """

        logger = logging.getLogger(logger_name)

        return logger


logging.setLoggerClass(CustomLogger)



import sys
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional, TypeVar

from pandas import DataFrame, concat, read_csv
from numpy import concatenate

from drive.log import CustomLogger
from drive.models import FileIndices, Genes

logger = CustomLogger.get_logger(__name__)

# we are going to create two exception class for the vertex

T = TypeVar("T", bound="IbdFilter")


@dataclass
class IbdFilter:
    ibd_file: Iterator[DataFrame]
    indices: FileIndices
    target_gene: Genes
    filter: Optional[Callable] = None
    ibd_vs: DataFrame = field(default_factory=DataFrame)
    ibd_pd: DataFrame = field(default_factory=DataFrame)
    hapid_map: Dict[str, int] = field(default_factory=dict)
    all_haplotypes: List[str] = field(default_factory=list)
    haplotype_id: int = 0

    @classmethod
    def load_file(
        cls,
        ibd_file: Path,
        indices: FileIndices,
        target_gene: Genes,
    ) -> T:
        """Factory method that returns the IBDFilter model
        This method makes sure that the ibd file exists

        Parameters
        ----------
        ibd_file : Path
            Path object containing the filepath for the ibd
            file from hapibd, iLASH, etc...

        indices: FileIndices
            Object that has all the indices for the necessary
            columns in the ibd file. This object also has a
            get_haplotype_ids method that will form the
            haplotype ibd.

        target_gene : Genes
            namedtuple that has attributes for the
            chromosome, the gene start position, and the
            gene end position.

        Returns
        -------
        IbdFilter
            returns an initialized IbdFilter object

        Raises
        ------
        FileNotFoundError
            raises an error if the file doesn't exist
        """
        logger.verbose(f"Reading in the ibd input file at {ibd_file}")

        if not ibd_file.is_file():
            raise FileNotFoundError(f"The file, {ibd_file}, was not found")

        # we need to make sure that the id columns read in as strings no matter what
        input_file_chunks = read_csv(
            ibd_file,
            sep="\t",
            header=None,
            chunksize=100_100,
            dtype={indices.id1_indx: str, indices.id2_indx: str},
        )

        return cls(input_file_chunks, indices, target_gene)

    def _generate_map(self, chunk_data: DataFrame) -> None:
        """Method that will generate the dictionary that maps hapibd to integers

        Parameters
        ----------
        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is determined by the
            chunksize argument to pd.read_csv. This value is currently set to 100,000.
        """
        haplotypes = chunk_data.values.ravel()

        logger.verbose(f"identified {len(haplotypes)} haplotypes in this chunk.")

        # iterate over each haplotype and add it to the
        # dictionary if the value is not present
        for value in haplotypes:
            key_value = self.hapid_map.setdefault(value, self.haplotype_id)
            if key_value == self.haplotype_id:
                self.haplotype_id += 1

    def _map_grids(self, data_chunk: DataFrame) -> None:
        """Function responsible for creating two new columns that
        have the haplotype id numbers.

        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is
            determined by the chunksize argument to
            pd.read_csv. This value is currently set to 100,000.
        """
        # we are going to map the haplotype id to integers in a new
        # column. this needs to be done for both hapid1 an hapid2
        data_chunk.loc[:, "idnum1"] = data_chunk["hapid1"].map(self.hapid_map)

        data_chunk.loc[:, "idnum2"] = data_chunk["hapid2"].map(self.hapid_map)

    def _contains_filter(self, data_chunk: DataFrame, min_cm: int) -> DataFrame:
        """Method that will filter the ibd file on four conditions: Chromosome number is the same, segment start position is <= target start position, segment end position is >= to the start position, and the size of the segment is >= to the minimum centimorgan threshold.

        Parameters
        ----------
        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is
            determined by the chunksize argument to
            pd.read_csv. This value is currently set to 100,000.

        min_cm : int
            centimorgan threshold

        Returns
        -------
        pd.DataFrame
            returns the filtered dataframe

        Raises
        ------
        ValueError
            raises a ValueError if the target chromosome number is not
            found within the provided IBD file. This situation will
            lead to a error later in the program which is why the
            exception is raised. It is assumed to be due the user
            providing the incorrect file by accident
        """  # noqa: E501
        # we are going to first make sure that the ibd file is for the
        # right chromosome. If the target_gene chromosome number is not
        # found in the file then a ValueError is raised.

        # sometimes build 38 have chr# instead of # in the file so we should also
        # check for that
        chromo_str = f"chr{self.target_gene.chr}"

        if (
            self.target_gene.chr not in data_chunk[self.indices.chr_indx].values
            and chromo_str not in data_chunk[self.indices.chr_indx].values
        ):
            error_msg = f"Expected the value of the chromosome column in the ibd file to be {self.target_gene.chr} or {chromo_str}. This value was not found in the column. Please ensure that you selected the proper IBD file for chromosome {self.target_gene.chr} before re-running DRIVE."  # noqa: E501

            logger.critical(error_msg)

            raise ValueError(error_msg)
        # We are going to filter the data and then make a copy
        # of it to return so that we don't get the
        # SettingWithCopyWarning

        return data_chunk[
            (data_chunk[self.indices.chr_indx] == self.target_gene.chr)
            & (data_chunk[self.indices.str_indx] <= self.target_gene.start)
            & (data_chunk[self.indices.end_indx] >= self.target_gene.end)
            & (data_chunk[self.indices.cM_indx] >= min_cm)
        ].copy()

    def _overlaps_filter(self, data_chunk: DataFrame, min_cm: int) -> DataFrame:
        """Method that will filter the ibd file on four conditions: Chromosome number is the same, segment start position is <= target start position, segment end position is >= to the start position, and the size of the segment is >= to the minimum centimorgan threshold.

        Parameters
        ----------
        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is
            determined by the chunksize argument to
            pd.read_csv. This value is currently set to 100,000.

        min_cm : int
            centimorgan threshold

        Returns
        -------
        pd.DataFrame
            returns the filtered dataframe

        Raises
        ------
        ValueError
            raises a ValueError if the target chromosome number is not
            found within the provided IBD file. This situation will
            lead to a error later in the program which is why the
            exception is raised. It is assumed to be due the user
            providing the incorrect file by accident
        """  # noqa: E501
        # we are going to first make sure that the ibd file is for the
        # right chromosome. If the target_gene chromosome number is not
        # found in the file then a ValueError is raised.
        if self.target_gene.chr not in data_chunk[self.indices.chr_indx].values:
            error_msg = f"Expected the value of the chromosome column in the ibd file to be {self.target_gene.chr}. This value was not found in the column. Please ensure that you selected the proper IBD file for chromosome {self.target_gene.chr} before re-running DRIVE."  # noqa: E501

            logger.critical(error_msg)

            raise ValueError(error_msg)
        # We are going to filter the data and then make a copy
        # of it to return so that we don't get the
        # SettingWithCopyWarning
        return data_chunk[
            (
                (
                    (data_chunk[self.indices.str_indx] <= int(self.target_gene.start))
                    & (data_chunk[self.indices.end_indx] >= int(self.target_gene.start))
                )
                | (
                    (data_chunk[self.indices.str_indx] >= int(self.target_gene.start))
                    & (data_chunk[self.indices.end_indx] <= int(self.target_gene.end))
                )
                | (
                    (data_chunk[self.indices.str_indx] <= int(self.target_gene.end))
                    & (data_chunk[self.indices.end_indx] >= int(self.target_gene.end))
                )
            )
            & (data_chunk[self.indices.cM_indx] >= min_cm)
        ].copy()

    def set_filter(self, filter_option: str) -> None:
        """Method to determine how the user wishes to filter the IBD segments file

        Parameters
        ----------
        filter_option : str
            string that represents the user's choice for how to filter the ibd segments.
            If the user chooses 'contains' the only segments that contain the entire
            region are kept. If the user chooses 'overlaps' then segments that overlap
            at all with the target region are kept."""

        if filter_option == "contains":
            logger.info("Identifying IBD segments that contain the target region")
            self.filter = self._contains_filter
        elif filter_option == "overlaps":
            logger.info("Identifying IBD segments that overlap the target region")
            self.filter = self._overlaps_filter
        else:
            logger.critical(
                "Non-recognized filter option selected. Allowed values are 'contains' and 'overlaps'. Exiting program now..."  # noqa: E501
            )
            sys.exit(0)

    def _remove_dups(self, data: DataFrame) -> DataFrame:
        """Filters out rows where the haplotype ids are the
        same

        Parameters
        ----------
        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is
            determined by the chunksize argument to
            pd.read_csv. This value is currently set to 100,000.

        Returns
        -------
        DataFrame
            returns a dataframe where we make sure that hapibd1 does != hapibd2

        Raises
        ------
        KeyError
            raises a key error if hapid1 or hapid2 are not columns in the dataframe
        """
        try:
            return data[data["hapid1"] != data["hapid2"]]
        except KeyError as e:
            logger.critical(
                f"Expected the keys hapid1 and hapid2 to be in the dataframe. Instead the only keys were: {', '.join(data.columns)}"  # noqa: E501
            )
            raise e(
                f"Expected the keys hapid1 and hapid2 to be in the dataframe. Instead the only keys were: {', '.join(data.columns)}"  # noqa: E501
            )

    def _generate_vertices(self, data_chunk: DataFrame) -> None:
        """Method that will generate the vertices dataframe which just has the
        columns idnum, hapID, and IID

        Parameters
        ----------
        data_chunk : pd.DataFrame
            chunk of the ibdfile. The size of this chunk is
            determined by the chunksize argument to
            pd.read_csv. This value is currently set to 100,000.
        """
        id1_df = data_chunk[["idnum1", "hapid1", self.indices.id1_indx]].rename(
            columns={"idnum1": "idnum", "hapid1": "hapID", 0: "IID"}
        )

        self.ibd_vs = concat([self.ibd_vs, id1_df])

        id2_df = data_chunk[["idnum2", "hapid2", self.indices.id2_indx]].rename(
            columns={"idnum2": "idnum", "hapid2": "hapID", 2: "IID"}
        )

        self.ibd_vs = concat([self.ibd_vs, id2_df])

    def _filter_for_cohort(
        self, chunk: DataFrame, cohort_ids: Optional[List[str]] = None
    ) -> DataFrame:
        """filter cohort chunk to individuals in the cohort
        list

        Parameters
        ----------
        chunk : DataFrame
            chunk of pandas dataframe that has information about the shared
            pairwise IBD segment.

        cohort_ids : List[str]
            Lists of ids that make up the cohort. The ibd_file will be filtered to only this list.

        Returns
        -------
        DataFrame
            returns the filtered pandas dataframe
        """  # noqa: E501
        # if no cohort ids were provided then we just return the chunk, otherwise we
        # filter the dataframe for where id1 and id2 are in the cohort
        if not cohort_ids:
            return chunk
        else:
            return chunk[
                (chunk[self.indices.id1_indx].isin(cohort_ids))
                & (chunk[self.indices.id2_indx].isin(cohort_ids))
            ]

    def _check_for_no_shared_segments(ibd_pd: DataFrame, ibd_vs: DataFrame) -> None:
        """Check to ensure that there were shared IBD segments
        found based on the input conditions

        Parameters
        ----------
        ibd_pd : DataFrame
            dataframe that has all the pairwise shared
            segments that satisfy the input conditions

        ibd_vs : DataFrame
            dataframe that has the information about each
            haplotype and individual that are in ibd_pd

        Raises
        ------
        ValueError
            raises a ValueError if the dataframes are empty
            because they cannot be empty or other steps of the
            program will fail.
        """
        if ibd_pd.empty:
            logger.critical(
                "There were no shared IBD segments that satisfied the provided input conditions"  # noqa: E501
            )
            raise ValueError(
                "There were no shared IBD segments that satisfied the provided input conditions. Please check to make sure that there is no typo in the target region or that the correct ibd input file was provided."  # noqa: E501
            )
        elif ibd_vs.empty:
            logger.critical(
                "There were no haplotype ids identified in the filtering step."
            )
            raise ValueError(
                "There were no haplotype ids identified in the filtering step."
            )
        else:
            logger.debug(
                f"Identified {ibd_pd.shape[0]} shared ibd segments with {ibd_vs.shape[0]} unique haplotypes"  # noqa: E501
            )

    def _check_empty_dataframes(self) -> None:
        """Check if the provided dataframe is empty. If it is
        then it raises an error and exits the program"""
        if self.ibd_pd.empty:
            logger.info(
                "No individuals from the analysis cohort share an IBD segment across the provided target region. Please ensure that the target region is correct. Exiting program now."  # noqa: E501
            )
            sys.exit(0)

    def preprocess(
        self,
        min_centimorgan: int,
        cohort_ids: Optional[List[str]] = None,
    ) -> None:
        """Method that will filter the ibd file.

        Parameters
        ----------
        min_centimorgan : int
            Minimum segment threshold that is used to filter
            the ibd file. Program only keeps segments that
            are greater than or equal to the threshold.

        cohort_ids : List[str]
            Lists of ids that make up the cohort. The ibd_file
            will be filtered to only this list.
        """
        # getting the start time for when the program begines to read in the ibd file
        start_time = datetime.now()

        for chunk in self.ibd_file:
            cohort_restricted_chunk = self._filter_for_cohort(chunk, cohort_ids)

            logger.debug(
                f"Identified {cohort_restricted_chunk.shape[0]} pairs in this chunk"
            )  # noqa: E501

            if cohort_restricted_chunk.empty:
                continue

            size_filtered_chunk = self.filter(cohort_restricted_chunk, min_centimorgan)

            logger.debug(
                f"{size_filtered_chunk.shape[0]} pairs remaining after filtering for the loci of interest with a {min_centimorgan} minimum shared segment threshold"  # noqa: E501
            )

            if not size_filtered_chunk.empty:
                # We have to add two column with the haplotype ids
                self.indices.get_haplotype_id(
                    size_filtered_chunk,
                    self.indices.id1_indx,
                    self.indices.hap1_indx,
                    "hapid1",
                )

                self.indices.get_haplotype_id(
                    size_filtered_chunk,
                    self.indices.id2_indx,
                    self.indices.hap2_indx,
                    "hapid2",
                )
                # We then need to make sure that there are no
                # duplicates in the dataframe
                removed_dups = self._remove_dups(size_filtered_chunk)

                # We need to update the mappings for the grids
                self._generate_map(removed_dups[["hapid1", "hapid2"]])

                self._map_grids(removed_dups)
                # concat the filtered dataframe with the ibd_pd
                # attribute to get all the edges in the graph
                self.ibd_pd = concat([self.ibd_pd, removed_dups])

                self._generate_vertices(removed_dups)

        self._check_empty_dataframes()

        self.ibd_pd.reset_index(drop=True, inplace=True)

        ids_in_ibd_pd = len(
            concatenate([self.ibd_pd.idnum1.unique(), self.ibd_pd.idnum2.unique()])
        )

        logger.info(
            f"{ids_in_ibd_pd} out of the {len(cohort_ids)} ids provided were found within the IBD data file"
        )

        self.ibd_vs = self.ibd_vs.drop_duplicates().sort_values(by="idnum")

        # We are going to print out how long it took to read in the ibd file
        # if the user puts the program in verbose mod
        end_time = datetime.now()

        logger.verbose(
            f"Analysis finished at {end_time}. Total runtime: {end_time - start_time}"
        )



from typing import Dict, List

from pandas import DataFrame, Series


def _identify_carriers_indx(
    carriers_series: Series, return_dict: Dict[str, List[str]]
) -> None:
    """Function that will insert the key, value pair of the phenotype and the carriers
    into the dictionary
    Parameters
    ----------
    carriers_series : pd.Series
        row of dataframe that is used in the apply function in
        the generate_carrier_list function

    return_dict : dict[str, list[str]]
        dictionary where the phecodes will be the keys and the values will be a list of
        indices indicating which grids are carriers
    """

    return_dict[carriers_series.name] = carriers_series[
        carriers_series == 1
    ].index.tolist()


def _identify_carriers_indx(
    carriers_series: Series, return_dict: Dict[str, List[str]]
) -> None:
    """Function that will insert the key, value pair of the phenotype and the carriers
    into the dictionary

    Parameters
    ----------
    carriers_series : pd.Series
        row of dataframe that is used in the apply function in
        the generate_carrier_list function

    return_dict : dict[str, list[str]]
        dictionary where the phecodes will be the keys and the values will be a list of
        indices indicating which grids are carriers
    """

    return_dict[carriers_series.name] = carriers_series[
        carriers_series == "NA"
    ].index.tolist()


def generate_carrier_dict(carriers_matrix: DataFrame) -> Dict[str, List[str]]:
    """Function that will take the carriers_pheno_matrix and generate a dictionary that
    has the list of indices for each carrier

    Parameters
    ----------
    carriers_matrix : pd.DataFrame
        dataframe where the columns are the phecodes and have 0's or 1's for whether or
        not they have the phecodes

    Returns
    -------
    Tuple[dict[str, list[str]], dict[str, List[str]]
        dictionary where the keys are phecodes and the values are list of integers
    """
    return_dict = {}

    new_indx_df = carriers_matrix.set_index("grids")

    # iterating over each phenotype which starts with the
    # second column
    new_indx_df.apply(lambda x: _identify_carriers_indx(x, return_dict), axis=0)

    return return_dict



import argparse
import sys
from pathlib import Path


class CheckInputExist(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs) -> None:
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(CheckInputExist, self).__init__(option_strings, dest, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Path,
        option_string: str = None,
    ) -> None:
        if values.exists():
            setattr(namespace, self.dest, values)
        else:
            print(
                f"ERROR: The file, {values}, was not found. Please make sure that there is not a typo in the file name."
            )
            sys.exit(1)


# class CheckJsonPath(argparse.Action):

#     def __init__(self, option_strings, dest, nargs=None, **kwargs) -> None:
#         if nargs is not None:
#             raise ValueError("nargs not allowed")
#         super(CheckJsonPath, self).__init__(option_strings, dest, **kwargs)


#     def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: Path, option_string: str =None) -> None:
#         print(f"This is the content in values: {values}")
#         if values:
#             setattr(namespace, self.dest, values)
#         else:
#             src_dir = Path(__file__).parent.parent.parent

#             config_path = src_dir / "config.json"

#             if not config_path.exists():
#                 raise FileNotFoundError(
#                     f"Expected the user to either pass a configuration file path or for the config.json file to be present in the program root directory at {config_path}."  # noqa: E501
#                 )

#             setattr(namespace, self.dest, config_path)



import itertools
import logging
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple

import igraph as ig
from pandas import DataFrame

from drive.log import CustomLogger
from drive.models import Filter, Network, Network_Interface

# creating a logger
logger: logging.Logger = CustomLogger.get_logger(__name__)

# Create a generic variable that can represents the class from the


@dataclass
class ClusterHandler:
    """Class responsible for performing the cluster on the network objects"""

    minimum_connected_thres: float
    max_network_size: int
    max_rechecks: int
    random_walk_step_size: int
    min_cluster_size: int
    segment_dist_threshold: int
    hub_threshold: float
    haplotype_mappings: Dict[str, int]
    recluster: bool
    check_times: int = 0
    recheck_clsts: Dict[int, List[Network_Interface]] = field(default_factory=dict)
    final_clusters: List[Network_Interface] = field(default_factory=list)

    @staticmethod
    def generate_graph(
        ibd_edges: DataFrame,
        ibd_vertices: Optional[DataFrame] = None,
    ) -> ig.Graph:
        """Method that will be responsible for creating the graph
        used in the network analysis

        Parameters
        ----------
        ibd_edges : DataFrame
            DataFrame that has the edges of the graph with the length
            of the edges.

        ibd_vertices : Optional[DataFrame]
            DataFrame that has information for each vertice in the
            graph. This value will be none when we are redoing the clustering
        """
        if ibd_vertices is not None:
            logger.debug("Generating graph with vertex labels.")
            return ig.Graph.DataFrame(
                ibd_edges, directed=False, vertices=ibd_vertices, use_vids=False
            )

        else:
            logger.debug(
                "No vertex metadata provided. Vertex ids will be nonnegative integers"
            )
            # return ig.Graph.DataFrame(ibd_edges, directed=False, use_vids=False)
            return ig.Graph.DataFrame(ibd_edges, directed=False)

    def random_walk(self, graph: ig.Graph) -> ig.VertexClustering:
        """Method used to perform the random walk from igraph.community_walktrap

        Parameters
        ----------
        graph : ig.Graph
            graph object created by ig.Graph.DataFrame

        Returns
        -------
        ig.VertexClustering
            result of the random walk cluster. This object has
            information about clusters and membership
        """
        logger.debug(
            f"Performing the community walktrap algorithm with a random walk step size of: {self.random_walk_step_size}"
        )

        ibd_walktrap = ig.Graph.community_walktrap(
            graph, weights="cm", steps=self.random_walk_step_size
        )

        random_walk_clusters = ibd_walktrap.as_clustering()

        logger.verbose(random_walk_clusters.summary())

        return random_walk_clusters

    def filter_cluster_size(self, random_walk_clusters_sizes: List[int]) -> List[int]:
        """Method to filter networks that are smaller than the min_cluster_size from
        the analysis

        Parameters
        ----------
        random_walk_clusters_sizes : List[int]
            size of each cluster from the random walk results

        Returns
        -------
        List[int]
            returns a list of integers where each integer
            represents a cluster id. Each cluster in the
            network will be >= to the min_cluster_size
            attribute.
        """
        return [
            i
            for i, v in enumerate(random_walk_clusters_sizes)
            if v > self.min_cluster_size
        ]

    @staticmethod
    def _gather_members(
        random_walk_members: List[int], clst_id: int, graph: ig.Graph
    ) -> Tuple[List[int], List[int]]:
        """Generate a list of individuals ids in the network

        Parameters
        ----------
        random_walk_members : List[int]
            list of all members from the random walk results

        clst_id : int
            id for the cluster. This value is used to pull out members
            that belong to the cluster

        graph : ig.Graph
            Graph object formed by ig.Graph.DataFrame

        Returns
        -------
        Tuple[List[int], List[int]]
            returns a list of ids of individuals in the network and then
            a list of vertex ids. The individual ids are just the index
            of the element in the membership list and the vertex ids are
            the list of ids provided by nme label in the vs() property.
        """
        member_list = []
        # this list has the ids. It is sometimes the same as the
        # member_list but it will not be the same in the redo_networks
        # graph
        vertex_ids = []

        for member_id, assigned_clst_id in enumerate(random_walk_members):
            if assigned_clst_id == clst_id:
                member_list.append(graph.vs()[member_id]["name"])
                vertex_ids.append(member_id)

        return member_list, vertex_ids

    @staticmethod
    def _determine_true_positive_edges(
        member_list: List[int], clst_id: int, random_walk_results: ig.VertexClustering
    ) -> Tuple[int, float]:
        """determining the number of true positive edges

        Parameters
        ----------
        member_list : List[int]
            list of ids within the specific network

        clst_id : int
            id for the original cluster

        random_walk_results : ig.VertexClustering
            vertexClustering object returned after the random
            walk that has the different clusters

        Returns
        -------
        Tuple[int, float]
            returns a tuple where the first element is the
            number of edges in the graph and the second
            element is the ratio of actually edges in the
            graph compared to the theoretical maximum number
            of edges in the graph.
        """
        # getting the total number of edges possible
        theoretical_edge_count = len(list(itertools.combinations(member_list, 2)))

        # Getting the number of edges within the graph and saving it
        # as a dictionary key, 'true_positive_n'
        cluster_edge_count = len(random_walk_results.subgraph(clst_id).get_edgelist())

        return cluster_edge_count, cluster_edge_count / theoretical_edge_count

    def _determine_false_positive_edges(
        graph: ig.Graph, vertex_list: List[int]
    ) -> Tuple[int, List[int]]:
        """determine the number of false positive edges

        Parameters
        ----------
        graph : ig.Graph
            graph object returned from ig.Graph.DataFrame

        vertex_list : List[int]
            list of vertex ids within the specific network

        Returns
        -------
        Tuple[int, List[int]]
            returns a tuple where the first element is the
            number of edges in the graph and the second
            element is a list of false positive edges.
        """
        all_edge = set([])

        for mem in vertex_list:
            all_edge = all_edge.union(set(graph.incident(mem)))

        false_negative_edges = list(
            all_edge.difference(
                list(
                    graph.get_eids(
                        pairs=list(itertools.combinations(vertex_list, 2)),
                        directed=False,
                        error=False,
                    )
                )
            )
        )
        return len(false_negative_edges), false_negative_edges

    def _map_ids_back_to_haplotypes(
        self, members: List[int]
    ) -> Tuple[List[str], Set[str]]:
        """remap the haplotype integer ids back to the haplotype
        strings for the output file

        Parameters
        ----------
        members : List[int]
            list of integers that represent the name of each vertex in the network
            corresponding the the graph

        Returns
        -------
        Tuple[List[str], List[str]]
            returns a list of haplotype id strings in the network. These
            strings include the phase number. Also returns a list of ids
            within the networks. This will be the same as the haplotype
            id without the phase number
        """
        haplotypes = [self.haplotype_mappings[value] for value in members]

        member_ids = {value[:-2] for value in haplotypes}

        return haplotypes, member_ids

    def gather_cluster_info(
        self,
        graph: ig.Graph,
        cluster_ids: List[int],
        random_walk_clusters: ig.VertexClustering,
        parent_cluster_id: Optional[str] = None,
    ) -> None:
        """Method for getting the information about membership,
        true.positive, false.positives, etc... from the random
        walk

        Parameters
        ----------
        graph : ig.Graph
            Graph object generated by ig.Graph.DataFrame

        cluster_ids : List[int]
            list of integers for each cluster id

        random_walk_clusters : ig.VertexClustering
            result from performing the random walk.

        parent_cluster_id : Optional[str]
            id of the original cluster that is now being broken up. Child
            cluster ids will take the form parent_id.child_id
        """

        for clst_id in cluster_ids:
            # We need to form the appropriate id if the cluster has a
            # parent otherwise they get the value of the clst_id argument
            if parent_cluster_id:
                clst_name = f"{parent_cluster_id}.{clst_id}"
            else:
                clst_name = f"{clst_id}"

            # We are going to get the vertex id and member id of each
            # graph
            member_list, vertex_ids = ClusterHandler._gather_members(
                random_walk_clusters.membership, clst_id, graph
            )

            # Next we get the number of edges/ ratio of actual edges to
            # the potential edges
            (
                true_pos_count,
                true_pos_ratio,
            ) = ClusterHandler._determine_true_positive_edges(
                member_list, clst_id, random_walk_clusters
            )
            # next we determine the number of false positive edges
            (
                false_neg_count,
                false_neg_list,
            ) = ClusterHandler._determine_false_positive_edges(graph, vertex_ids)

            # If the graph is too sparse and it is too large and the max
            # number of rechecks has not been reached then we will put
            # the network into a recluster dictionary. Otherwise it is
            # added to the final_clst list
            if (
                self.check_times < self.max_rechecks
                and true_pos_ratio < self.minimum_connected_thres
                and len(member_list) > self.max_network_size
                and self.recluster
            ):
                # We can put all of this information into a network class. Here the
                # member list will still be in integers
                network = Network(
                    clst_name,
                    true_pos_count,
                    true_pos_ratio,
                    false_neg_list,
                    false_neg_count,
                    member_list,
                    vertex_ids,
                )
                # debug statement if we want to see the members and the haplotypes
                logger.debug(f"members: {member_list}\nhaplotypes: {vertex_ids}")

                # we are going to append the network to the list of networks that needs
                # to be rechecked
                self.recheck_clsts.setdefault(self.check_times, []).append(network)

            else:
                # we need to convert the integer ids back to strings
                haplotype_ids, member_ids = self._map_ids_back_to_haplotypes(
                    member_list
                )

                network = Network(
                    clst_name,
                    true_pos_count,
                    true_pos_ratio,
                    false_neg_list,
                    false_neg_count,
                    member_ids,
                    haplotype_ids,
                )
                # logger.info(f"members: {member_ids}\nhaplotypes: {haplotype_ids}")

                self.final_clusters.append(network)

    def redo_clustering(
        self, network: Network_Interface, ibd_pd: DataFrame, ibd_vs: DataFrame
    ) -> None:
        """Method that will redo the clustering, if the
        networks were too large or did not show a high degree
        of connectedness

        Parameters
        ----------
        network : Network_InterFace
            object that represents each cluster. These objects have information
            about the cluster id, number and ratio of edges, true_positive_percent,
            false_negative_edges, false_negative_count

        ibd_pd : pd.DataFrame
            DataFrame that has information about the edges that a pair shares
        """
        # pulling the id from the original cluster
        original_id = network.clst_id

        # filters for the specific cluster
        redopd = ibd_pd[
            (ibd_pd["idnum1"].isin(network.haplotypes))
            & (ibd_pd["idnum2"].isin(network.haplotypes))
        ]

        redo_vs = ibd_vs[ibd_vs.idnum.isin(network.haplotypes)]

        # If the redopd or redo_vs is empty it causes strange behavior and the code will
        # usually fail. The desired behavior is for the program to tell teh user that
        # the graph could not be constructed and then for it to move on.
        if not redopd.empty and not redo_vs.empty:
            # We are going to generate a new Networks object using the redo graph
            redo_networks = ClusterHandler.generate_graph(redopd, redo_vs)
            # redo_networks = ClusterHandler.generate_graph(redopd)
            # performing the random walk
            redo_walktrap_clusters = self.random_walk(redo_networks)

            # If only one cluster is found
            if len(redo_walktrap_clusters.sizes()) == 1:
                # creates an empty dataframe with these columns
                clst_conn = DataFrame(columns=["idnum", "conn", "conn.N", "TP"])
                # iterate over each member id
                # for idnum in network.haplotypes:
                for idnum in network.members:
                    conn = sum(
                        list(
                            map(
                                lambda x: 1 / x,
                                redopd.loc[
                                    (redopd["idnum1"] == idnum)
                                    | (redopd["idnum2"] == idnum)
                                ]["cm"],
                            )
                        )
                    )
                    conn_idnum = list(
                        redopd.loc[(redopd["idnum1"] == idnum)]["idnum2"]
                    ) + list(redopd.loc[(redopd["idnum2"] == idnum)]["idnum1"])
                    conn_tp = len(
                        redopd.loc[
                            redopd["idnum1"].isin(conn_idnum)
                            & redopd["idnum2"].isin(conn_idnum)
                        ].index
                    )
                    # assert 1 == 0
                    if len(conn_idnum) == 1:
                        connTP = 1
                    else:
                        try:
                            connTP = conn_tp / (
                                len(conn_idnum) * (len(conn_idnum) - 1) / 2
                            )
                        except ZeroDivisionError:
                            raise ZeroDivisionError(
                                f"There was a zero division error encountered when looking at the network with the id {idnum}"  # noqa: E501
                            )  # noqa: E501

                    clst_conn.loc[idnum] = [idnum, conn, len(conn_idnum), connTP]
                rmID = list(
                    clst_conn.loc[
                        (
                            clst_conn["conn.N"]
                            > (self.segment_dist_threshold * len(network.members))
                        )
                        & (clst_conn["TP"] < self.minimum_connected_thres)
                        & (
                            clst_conn["conn"]
                            > sorted(clst_conn["conn"], reverse=True)[
                                int(self.hub_threshold * len(network.members))
                            ]
                        )
                    ]["idnum"]
                )

                redopd = redopd.loc[
                    (~redopd["idnum1"].isin(rmID)) & (~redopd["idnum2"].isin(rmID))
                ]
                redo_graph = self.generate_graph(
                    redopd,
                )
                # redo_g = ig.Graph.DataFrame(redopd, directed=False)
                redo_walktrap_clusters = self.random_walk(redo_graph)
                # redo_walktrap = ig.Graph.community_walktrap(
                #     redo_g, weights="cm", steps=self.random_walk_step_size
                # )
                # redo_walktrap_clusters = redo_walktrap.as_clustering()

            # Filter to the clusters that are llarger than the minimum size
            allclst = self.filter_cluster_size(redo_walktrap_clusters.sizes())

            self.gather_cluster_info(
                redo_networks, allclst, redo_walktrap_clusters, original_id
            )
        else:
            logger.debug(
                f"A graph was not able to be generated when we attempted to recluster the network: {original_id}. This error probably indicates that there were There were none of the {len(network.haplotypes)} individuals in that specific network that shared ibd segments with one another."
            )


def cluster(
    filter_obj: Filter,
    cluster_obj: ClusterHandler,
    centimorgan_indx: int,
) -> List[Network_Interface]:
    """Main function that will perform the clustering using igraph

    Parameters
    ----------
    filter_obj : Filter
        Filter object that has two attributes: ibd_pd and ibd_vs. These
        attributes are two dataframes that have information about the
        edges and information about the vertices.

    cluster_obj : ClusterHandler
        Object that contains information about how the random walk
        needs to be performed. It will use the construct networks and return those
        values in a list.

    min_network_size : int
        threshold so we can filter networks that are >= the threshold

        Returns
    """
    filter_obj.ibd_pd = filter_obj.ibd_pd.rename(columns={centimorgan_indx: "cm"})
    # filtering the edges dataframe to the correct columns
    ibd_pd = filter_obj.ibd_pd.loc[:, ["idnum1", "idnum2", "cm"]]

    ibd_vs = filter_obj.ibd_vs.reset_index(drop=True)

    # Generate the first pass networks
    network_graph = cluster_obj.generate_graph(
        ibd_pd,
        ibd_vs,
    )

    random_walk_results = cluster_obj.random_walk(network_graph)

    allclst = cluster_obj.filter_cluster_size(random_walk_results.sizes())

    cluster_obj.gather_cluster_info(network_graph, allclst, random_walk_results)

    while (
        cluster_obj.check_times < cluster_obj.max_rechecks
        and len(cluster_obj.recheck_clsts.get(cluster_obj.check_times, [])) > 0
    ):
        cluster_obj.check_times += 1
        logger.verbose(f"recheck: {cluster_obj.check_times}")

        _ = cluster_obj.recheck_clsts.setdefault(cluster_obj.check_times, [])

        for network in cluster_obj.recheck_clsts.get(cluster_obj.check_times - 1):
            cluster_obj.redo_clustering(network, ibd_pd, ibd_vs)
    # logginng the number of segments, haplotypes, and clusters
    # identified in the analysis
    logger.info(
        f"Identified {network_graph.ecount()} IBD segments from {network_graph.vcount()} haplotypes"  # noqa: E501
    )

    logger.info(f"Identified {len(cluster_obj.final_clusters)} IBD clusters")

    return cluster_obj.final_clusters



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Protocol, Set

from .networks import Network_Interface


class Data_Interface(Protocol):
    """Protocol defining what attributes the DataHolder Interface needs to have"""

    networks: List[Network_Interface]
    output_path: Path
    carriers: Dict[str, Dict[str, Set[str]]]
    phenotype_descriptions: Dict[str, Dict[str, str]]


@dataclass
class Data:
    """main class to hold the data from the network analysis and the different pvalues"""

    networks: List[Network_Interface]
    output_path: Path
    carriers: Dict[str, Dict[str, Set[str]]]
    phenotype_descriptions: Dict[str, Dict[str, str]]
